# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from climetlab.indexing.fieldset import FieldSet


class EmptySource(FieldSet):
    def ignore(self):
        # Used by multi-source
        return True

    def __len__(self):
        return 0

    def __repr__(self):
        return "EmptySource"


source = EmptySource
